/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*-
 *
 * Copyright (C) 2008-2013 Red Hat, Inc.
 *
 * Licensed under GPL version 2 or later.
 *
 * Author: David Zeuthen <zeuthen@gmail.com>
 */

#ifndef __LIB_GDU_TYPES_H__
#define __LIB_GDU_TYPES_H__

#include <gtk/gtk.h>
#include <udisks/udisks.h>

#include "libgduenums.h"

G_BEGIN_DECLS

G_END_DECLS

#endif /* __LIB_GDU_TYPES_H__ */
